/*
 * p0f - passive OS fingerprinting
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <pcap.h>
#include <arpa/inet.h>

#include "tcp.h"

#define MAXFPS 1000
#define FPBUF  120
#define INBUF  1024
#define TTLDW  30

char fps[MAXFPS][FPBUF];
int wss, wscale, mss, nop, ttl, df, sok,tmp,header_len=14,dupa;
struct in_addr sip;
struct bpf_program flt;
pcap_t *pt;

void lookup(void);

void parse(u_char *blabla, struct pcap_pkthdr *pph, u_char *packet) {
  struct iphdr *iph;
  struct tcphdr *tcph;
  int ilen=0, hlen=0,off;
  dupa=0;
  
  if (pph->len < header_len+sizeof(struct iphdr)+sizeof(struct tcphdr)) {
    return;
  }

  iph=(struct iphdr*) (packet+header_len);
  if ((iph->ihl>>4)!=4 || iph->protocol!=IPPROTO_TCP) {
    return;
  }

  ttl=iph->ttl;

  off=ntohs(iph->off);
  df=((off&IP_DF)!=0);
  sip.s_addr=iph->saddr;
  ilen= ( (iph->ihl&0x0f) );

  switch (ilen) {
    case 5: /* no options */
      tcph=(struct tcphdr *)(iph+1);
      break;
    default: /* parse ipoptions */
      if ((header_len+(ilen<<2)+sizeof(struct tcphdr)) > pph->len) {
	return;
      }
      tcph=(struct tcphdr *)(packet+header_len+(ilen<<2));
      break;
  }

  off=tcph->th_flags;
  if (!(off&TH_SYN)) return;
  if ((off&TH_ACK)) return;

  wscale=-1;
  mss=0;
  nop=0;
  sok=0;

  hlen=(tcph->th_off)*4;

  {
    void* opt_ptr;
    int opt;
    opt_ptr=(void*)tcph+sizeof(struct tcphdr);
    while (dupa<hlen) {
      opt=(int)(*(u_char*)(opt_ptr+dupa));
      dupa+=2;
      switch(opt) {
        case TCPOPT_EOL:
	  dupa=100000; break;
	case TCPOPT_MAXSEG:
  	  mss=EXTRACT_16BITS(opt_ptr+dupa);
  	  dupa+=2;
	  break;
        case TCPOPT_NOP:
  	  nop=1;
	  break;
	case TCPOPT_WSCALE:
  	  wscale=EXTRACT_16BITS(opt_ptr+dupa);
	  dupa+=2;
	  break;
	case TCPOPT_TIMESTAMP:
          dupa+=8;
	  break;
	case TCPOPT_SACKOK:
 	  sok=1;
	default: break;
      }
    }
  }
#if __BYTE_ORDER == __LITTLE_ENDIAN
  wss=htons(tcph->th_win);
#else
  wss=tcph->th_win;
#endif
  lookup();
  return;
}


void lookup(void) {
  int i=0,got=0,down=0;
  char buf[INBUF],*p;
pump:
  // To be removed soon. Compatibility.
  for (down=0;down<TTLDW;down++) {
    i=0;
    sprintf(buf,"%d:%d:%d:%d:%d:%d:%d:",wss,ttl+down,mss,df,wscale,sok,nop);
    while (fps[i][0]) {
      if (!strncmp(buf, fps[i], strlen(buf))) {
        got=1;
        p=strrchr(fps[i],':')+1;
        if (!strchr(p, '\n')) strcat(p, "\n");
        printf("%s [%d hops]: %s",inet_ntoa(sip),down+1,p);
        break;
      }
      i++;
    }
    if (got) break;
  }
  if (!got) if (wscale==-1) { wscale=0; goto pump; }
  if (!got) printf("%s: UNKNOWN [%d:%d:%d:%d:%d:%d:%d].\n",
            inet_ntoa(sip), wss, ttl, mss, df, wscale, sok, nop);
  fflush(0);
}


void load_fprints(void) {
  FILE *x;
  int i=0;
  char *p;
  x=fopen("p0f.fp", "r");
  if (!x) x=fopen("/etc/p0f.fp", "r");
  if (!x) {
    fprintf(stderr, "No OS fingerprint database (p0f.fp) found. Dumb mode on.\n");
    return;
  }
  while (fgets(fps[i],FPBUF-1,x)) {
    if ((p=strchr(fps[i],'#')))	*p=0;
    if (fps[i][0]) i++;
  }
  fprintf(stderr, "Loaded %d fingerprints.\n",i-1);
  fclose(x);
}


int main(int argc, char *argv[]) {
  char errbuf[PCAP_ERRBUF_SIZE];
  fprintf(stderr, "p0f 1.0 - passive os fingerprinting (c) <lcamtuf@tpi.pl>\n");
  if ((pt=pcap_open_live(pcap_lookupdev(errbuf),100,1,100,errbuf))==NULL) {
    pcap_lookupdev(errbuf);
    fprintf(stderr, "pcap_open_live failed: %s\n", errbuf);
    exit(1);
  }
  load_fprints();

  if (pcap_compile(pt, &flt, (argc==2)?argv[1]:"", 1, 0)) {
    if (argc==2) {
      pcap_perror(pt,"pcap_compile");
      exit(1);
    }
  }
  
  if (argc==2) {
    fprintf(stderr,"Using custom filter rule: '%s'\n",argv[2]);
  }

  pcap_setfilter(pt, &flt);

  pcap_loop(pt,-1,(pcap_handler)&parse,(void*)0L);
  return 0; //not reached;>
}
